### Non-Life Insurance: Mathematics and Statistics
### Exercise 11.2
### sessionInfo() is at the end of the document

####################################################################################
### Load packages
####################################################################################

#install.packages("CASdatasets", repos = "http://cas.uqam.ca/pub/", type="source")
require(CASdatasets)
require(keras)
require(tensorflow)
require(plyr)

data("freMTPL2freq")
data <- freMTPL2freq
str(data)


####################################################################################
### Illustrate data
####################################################################################

data <- data[,c("IDpol", "ClaimNb", "Exposure", "VehPower", "VehAge", "DrivAge")]
str(data)

hist(data$Exposure, col="gray", ylab="number of policies", cex.lab=1.5, xlab="Exposures", cex.lab=1.5, cex.main=1.5, main=list(paste("histogram of Exposures (", nrow(data), " policies)", sep=""), cex=1.5))

#############################
plot_marginals <- function(xx, variable, color1="gray"){
   barplot(xx[,3], names.arg=xx[,1], ylim=c(0, max(xx[,3])), xlab=variable, col=color1, cex.lab=1.5, pch=19, ylab="Exposure", cex.main=1.5, main=list(paste("total volumes per ", variable, sep=""), cex=1.5))
              }

#############################
plot_marginal_frequency <- function(xx, variable, ylim1=0.2, color1="blue"){
   plot(x=c(1:nrow(xx)), y=xx[,4]/xx[,3], ylim=c(0,ylim1), col=color1, cex.lab=1.5, pch=19, cex=1.5, xaxt='n', xlab=variable, ylab="frequency", cex.main=1.5, main=list(paste("observed frequency per ", variable, sep=""), cex=1.5))
   lines(x=c(1:nrow(xx)), y=xx[,4]/xx[,3] + 2* sqrt(xx[,4])/xx[,3], col=color1, lty=2)
   lines(x=c(1:nrow(xx)), y=xx[,4]/xx[,3] - 2* sqrt(xx[,4])/xx[,3], col=color1, lty=2)
   points(x=c(1:nrow(xx)), y=xx[,4]/xx[,3], col="cyan", pch=20)
   axis(1,at=c(1:nrow(xx)),labels=xx[,1])
              }
#############################

dat <- data
dat$n <- 1

###1 vehicle power
xx <- ddply(dat, .(VehPower), summarise, n=sum(n), Exposure=sum(Exposure), Claims=sum(ClaimNb))
plot_marginals(xx, "vehicle power groups")
plot_marginal_frequency(xx, "vehicle power groups")


###2 vehicle age
xx <- ddply(dat, .(VehAge), summarise, n=sum(n), Exposure=sum(Exposure), Claims=sum(ClaimNb))
plot_marginals(xx, "vehicle age groups")
plot_marginal_frequency(xx, "vehicle age groups")


###3 driver age
xx <- ddply(dat, .(DrivAge), summarise, n=sum(n), Exposure=sum(Exposure), Claims=sum(ClaimNb))
plot_marginals(xx, "driver's age groups")
plot_marginal_frequency(xx, "driver's age groups")


####################################################################################
### Pre-process variables for GLM
####################################################################################

dat <- data
dat$VehPower <- relevel(as.factor(dat$VehPower), ref="6")
VehAgeCat <- cbind(c(0:100), c(1, rep(2,3), rep(3,2), rep(4,2), rep(5,3), rep(6,2), rep(7,2), rep(8,3), rep(9,83)))
dat$VehAge <- relevel(as.factor(VehAgeCat[dat$VehAge+1,2]), ref="2")
DrivAgeCat <- cbind(c(18:100), c(rep(1,21-18),rep(2,26-21),rep(3,31-26),rep(4,41-31),rep(5,51-41),rep(6,71-51),rep(7,101-71)))
dat$DrivAge <- relevel(as.factor(DrivAgeCat[dat$DrivAge-17,2]), ref="6")
str(dat)

####################################################################################
### Training set and test set
####################################################################################

set.seed(100)
train <- sample(1:nrow(data),round(0.5*nrow(data)))

# individual insurance policies
train.individual <- dat[train,]
test.individual  <- dat[-train,]

# aggregated in risk classes
train.aggregate <- ddply(dat[train,], .(VehPower,VehAge,DrivAge), summarise, ClaimNb = sum(ClaimNb), Exposure = sum(Exposure))[,c(4:5,1:3)]
test.aggregate <- ddply(dat[-train,], .(VehPower,VehAge,DrivAge), summarise, ClaimNb = sum(ClaimNb), Exposure = sum(Exposure))[,c(4:5,1:3)]

str(train.individual)
str(train.aggregate)

####################################################################################
### Exercise 11.2 a)
####################################################################################


### Apply GLM (on the training set)
d.glm <- glm(ClaimNb ~ VehPower + VehAge + DrivAge, data=train.aggregate, offset = log(Exposure), family = poisson())
summary(d.glm)
drop1(d.glm, , test="Chi")

d.glm2 <- glm(ClaimNb ~ VehPower + VehAge + DrivAge, data=train.individual, offset = log(Exposure), family = poisson())
drop1(d.glm2, , test="Chi")
summary(d.glm2)



### Deviance statistics on training set
predtrain <- predict(d.glm, train.aggregate, type = "response")
obstrain <- train.aggregate$ClaimNb
(Deviancetrain <- 2 * sum(log((obstrain/predtrain)^obstrain) - obstrain + predtrain))
d.glm$deviance   ### check deviance statistics on training set

### Deviance statistics on test set
predtestGLM <- predict(d.glm, test.aggregate, type = "response")
obstest <- test.aggregate$ClaimNb
(Deviancetest <- 2 * sum(log((obstest/predtestGLM)^obstest) - obstest + predtestGLM))


### Deviance statistics on training set
predtrain2 <- predict(d.glm, train.individual, type = "response")
obstrain2 <- train.individual$ClaimNb
(Deviancetrain2 <- 2 * sum(log((obstrain2/predtrain2)^obstrain2) - obstrain2 + predtrain2))
d.glm2$deviance   ### check deviance statistics on training set

### Deviance statistics on test set
predtestGLM2 <- predict(d.glm, test.individual, type = "response")
obstest2 <- test.individual$ClaimNb
(Deviancetest2 <- 2 * sum(log((obstest2/predtestGLM2)^obstest2) - obstest2 + predtestGLM2))



####################################################################################
#### Aggregated data
#### Exercise 11.2 b) and c)
####################################################################################


### Dummy coding
### Aggregated data
Ztrain <- model.matrix(data=train.aggregate, ClaimNb ~ VehPower + VehAge + DrivAge)
train.aggregate[,6:30] <- as.data.frame(Ztrain[,-1])
Ztest <- model.matrix(data=test.aggregate, ClaimNb ~ VehPower + VehAge + DrivAge)
test.aggregate[,6:30] <- as.data.frame(Ztest[,-1])
str(test.aggregate)

### Features
featlearn <- as.matrix(train.aggregate[,6:30])
feattest <- as.matrix(test.aggregate[,6:30])

### Volumes
vollearn <- as.matrix(train.aggregate$Exposure)
voltest <- as.matrix(test.aggregate$Exposure)

### Responses
resplearn <- as.matrix(train.aggregate$ClaimNb)
resptest <- as.matrix(test.aggregate$ClaimNb)

### Initial estimate
lambda0 <- sum(train.aggregate$ClaimNb)/sum(train.aggregate$Exposure)

### Keras model
k_clear_session()
seed1 <- 100
set.seed(seed1)
set_random_seed(seed1)
#
Design <- layer_input(shape = c(25), dtype = "float32", name = "Design")
Vol <- layer_input(shape = c(1), dtype = "float32", name = "Vol")
#
Network <- Design %>%
  layer_dense(units=20, activation="tanh", name="Layer1") %>%
  layer_dense(units=10, activation="tanh", name="Layer2") %>%
  layer_dense(units=1, activation=k_exp, name="Network",
              weights=list(array(0, dim=c(10,1)), array(log(lambda0), dim=c(1))))
#
Response <- list(Network, Vol) %>% layer_multiply()
#
model <- keras_model(inputs = c(Design, Vol), outputs = c(Response))
model %>% compile(loss = 'poisson', optimizer = 'nadam')
#
model


### Fit neural network model
epochs <- 100   ### c) 1000
fit <- model %>% fit(x = list(featlearn, vollearn), y = resplearn,
                     epochs = epochs, verbose = 0)
plot(fit)



### Deviance statistics on training set
predtrain <- as.vector(model %>% predict(list(featlearn, vollearn)))
obstrain <- train.aggregate$ClaimNb
c(Deviancetrain, 2 * sum(log((obstrain/predtrain)^obstrain) - obstrain + predtrain))

### Deviance statistics on test set
predtestNN <- as.vector(model %>% predict(list(feattest, voltest)))
obstest <- test.aggregate$ClaimNb
c(Deviancetest, 2 * sum(log((obstest/predtestNN)^obstest) - obstest + predtestNN))




####################################################################################
#### Individual data
#### Exercise 11.2 b) and c)
####################################################################################


### Dummy coding
### Individual data
Ztrain <- model.matrix(data=train.individual, ClaimNb ~ VehPower + VehAge + DrivAge)
train.individual[,7:31] <- as.data.frame(Ztrain[,-1])
Ztest <- model.matrix(data=test.individual, ClaimNb ~ VehPower + VehAge + DrivAge)
test.individual[,7:31] <- as.data.frame(Ztest[,-1])
str(test.individual)

### Features
featlearn <- as.matrix(train.individual[,7:31])
feattest <- as.matrix(test.individual[,7:31])

### Volumes
vollearn <- as.matrix(train.individual$Exposure)
voltest <- as.matrix(test.individual$Exposure)

### Responses
resplearn <- as.matrix(train.individual$ClaimNb)
resptest <- as.matrix(test.individual$ClaimNb)

### Initial estimate
lambda0 <- sum(train.individual$ClaimNb)/sum(train.individual$Exposure)

### Keras model
k_clear_session()
seed1 <- 100
set.seed(seed1)
set_random_seed(seed1)
#
Design <- layer_input(shape = c(25), dtype = "float32", name = "Design")
Vol <- layer_input(shape = c(1), dtype = "float32", name = "Vol")
#
Network <- Design %>%
  layer_dense(units=20, activation="tanh", name="Layer1") %>%
  layer_dense(units=10, activation="tanh", name="Layer2") %>%
  layer_dense(units=1, activation=k_exp, name="Network",
              weights=list(array(0, dim=c(10,1)), array(log(lambda0), dim=c(1))))
#
Response <- list(Network, Vol) %>% layer_multiply()
#
model <- keras_model(inputs = c(Design, Vol), outputs = c(Response))
model %>% compile(loss = 'poisson', optimizer = 'nadam')
#
model

file0 <- paste("callback", sep="")
CBs <- callback_model_checkpoint(file0, monitor = "val_loss", verbose = 0,  save_best_only = TRUE, save_weights_only = TRUE)


### Fit neural network model
epochs <- 200
fit <- model %>% fit(x = list(featlearn, vollearn), y = resplearn,
                     validation_split = 0.2,
                     batch_size = 5000,
                     callbacks=CBs,
                     epochs = epochs, verbose = 0)
plot(fit)
which.min(fit[[2]]$val_loss)


### Losses after all epochs
### Deviance statistics on training set
predtrain <- as.vector(model %>% predict(list(featlearn, vollearn), batch_size=10^7))
obstrain <- train.individual$ClaimNb
c(Deviancetrain2, 2 * sum(log((obstrain/predtrain)^obstrain) - obstrain + predtrain))

### Deviance statistics on test set
predtestNN <- as.vector(model %>% predict(list(feattest, voltest), batch_size=10^7))
obstest <- test.individual$ClaimNb
c(Deviancetest2, 2 * sum(log((obstest/predtestNN)^obstest) - obstest + predtestNN))


### Losses minimal validation loss fit
load_model_weights_hdf5(model, file0)
predtrain <- as.vector(model %>% predict(list(featlearn, vollearn), batch_size=10^7))
(insample <- c(Deviancetrain2, 2 * sum(log((obstrain/predtrain)^obstrain) - obstrain + predtrain)))

predtestNN <- as.vector(model %>% predict(list(feattest, voltest), batch_size=10^7))
(outofsample <- c(Deviancetest2, 2 * sum(log((obstest/predtestNN)^obstest) - obstest + predtestNN)))


####################################################################################
#### Individual data: continuous coding
#### Exercise 11.2 b) and c)
####################################################################################

## MinMaxScaling
dat <- data
dat$VehPowerNN  <- pmin(dat$VehPower,9)
dat$VehAgeNN  <- pmin(dat$VehAge,20)
dat$DrivAgeNN <- pmin(dat$DrivAge,80)

dat$VehPowerNN <- 2*(dat$VehPowerNN-min(dat$VehPowerNN))/(max(dat$VehPowerNN)-min(dat$VehPowerNN))-1
dat$VehAgeNN <- 2*(dat$VehAgeNN-min(dat$VehAgeNN))/(max(dat$VehAgeNN)-min(dat$VehAgeNN))-1
dat$DrivAgeNN <- 2*(dat$DrivAgeNN-min(dat$DrivAgeNN))/(max(dat$DrivAgeNN)-min(dat$DrivAgeNN))-1
str(dat)

####################################################################################
### Training set and test set
####################################################################################

set.seed(100)
train <- sample(1:nrow(data),round(0.5*nrow(data)))

# individual insurance policies
train.individual <- dat[train,]
test.individual  <- dat[-train,]

### Features
featlearn <- as.matrix(train.individual[,7:9])
feattest <- as.matrix(test.individual[,7:9])

### Volumes
vollearn <- as.matrix(train.individual$Exposure)
voltest <- as.matrix(test.individual$Exposure)

### Responses
resplearn <- as.matrix(train.individual$ClaimNb)
resptest <- as.matrix(test.individual$ClaimNb)

### Initial estimate
lambda0 <- sum(train.individual$ClaimNb)/sum(train.individual$Exposure)

### Keras model
k_clear_session()
seed1 <- 100
set.seed(seed1)
set_random_seed(seed1)
#
Design <- layer_input(shape = c(3), dtype = "float32", name = "Design")
Vol <- layer_input(shape = c(1), dtype = "float32", name = "Vol")
#
Network <- Design %>%
  layer_dense(units=20, activation="tanh", name="Layer1") %>%
  layer_dense(units=10, activation="tanh", name="Layer2") %>%
  layer_dense(units=1, activation=k_exp, name="Network",
              weights=list(array(0, dim=c(10,1)), array(log(lambda0), dim=c(1))))
#
Response <- list(Network, Vol) %>% layer_multiply()
#
model <- keras_model(inputs = c(Design, Vol), outputs = c(Response))
model %>% compile(loss = 'poisson', optimizer = 'nadam')
#
model

file0 <- paste("callback_cont", sep="")
CBs <- callback_model_checkpoint(file0, monitor = "val_loss", verbose = 0,  save_best_only = TRUE, save_weights_only = TRUE)


### Fit neural network model
epochs <- 300
fit <- model %>% fit(x = list(featlearn, vollearn), y = resplearn,
                     validation_split = 0.2,
                     batch_size = 5000,
                     callbacks=CBs,
                     epochs = epochs, verbose = 0)
plot(fit)
which.min(fit[[2]]$val_loss)


### Losses minimal validation loss fit
load_model_weights_hdf5(model, file0)
predtrain <- as.vector(model %>% predict(list(featlearn, vollearn), batch_size=10^7))
c(Deviancetrain2, 2 * sum(log((obstrain/predtrain)^obstrain) - obstrain + predtrain))
insample

predtestNN <- as.vector(model %>% predict(list(feattest, voltest), batch_size=10^7))
c(Deviancetest2, 2 * sum(log((obstest/predtestNN)^obstest) - obstest + predtestNN))
outofsample


